@@ -1,20 +1,24 @@
 """PodmanResource manager subclassed for Images."""
 
+import builtins
 import io
 import json
 import logging
+import os
 import urllib.parse
-from typing import Any, Dict, Generator, Iterator, List, Mapping, Optional, Union
+from typing import Any, Literal, Optional, Union
+from collections.abc import Iterator, Mapping, Generator
+from pathlib import Path
 import requests
 
 from podman import api
-from podman.api import Literal
-from podman.api.http_utils import encode_auth_header
+from podman.api.parse_utils import parse_repository
 from podman.domain.images import Image
 from podman.domain.images_build import BuildMixin
+from podman.domain.json_stream import json_stream
 from podman.domain.manager import Manager
 from podman.domain.registry_data import RegistryData
-from podman.errors import APIError, ImageNotFound
+from podman.errors import APIError, ImageNotFound, PodmanError
 
 try:
     from rich.progress import (
@@ -44,25 +48,28 @@ class ImagesManager(BuildMixin, Manager):
         response = self.client.get(f"/images/{key}/exists")
         return response.ok
 
-    def list(self, **kwargs) -> List[Image]:
+    def list(self, **kwargs) -> builtins.list[Image]:
         """Report on images.
 
         Keyword Args:
             name (str) – Only show images belonging to the repository name
             all (bool) – Show intermediate image layers. By default, these are filtered out.
-            filters (Mapping[str, Union[str, List[str]]) – Filters to be used on the image list.
+            filters (Mapping[str, Union[str, list[str]]) – Filters to be used on the image list.
                 Available filters:
 
                 - dangling (bool)
-                - label (Union[str, List[str]]): format either "key" or "key=value"
+                - label (Union[str, list[str]]): format either "key" or "key=value"
 
         Raises:
             APIError: when service returns an error
         """
+        filters = kwargs.get("filters", {}).copy()
+        if name := kwargs.get("name"):
+            filters["reference"] = name
+
         params = {
             "all": kwargs.get("all"),
-            "name": kwargs.get("name"),
-            "filters": api.prepare_filters(kwargs.get("filters")),
+            "filters": api.prepare_filters(filters=filters),
         }
         response = self.client.get("/images/json", params=params)
         if response.status_code == requests.codes.not_found:
@@ -113,50 +120,90 @@ class ImagesManager(BuildMixin, Manager):
             collection=self,
         )
 
-    def load(self, data: bytes) -> Generator[Image, None, None]:
+    def load(
+        self, data: Optional[bytes] = None, file_path: Optional[os.PathLike] = None
+    ) -> Generator[Image, None, None]:
         """Restore an image previously saved.
 
         Args:
             data: Image to be loaded in tarball format.
+            file_path: Path of the Tarball.
+                       It works with both str and Path-like objects
 
         Raises:
-            APIError: when service returns an error
+            APIError: When service returns an error.
+            PodmanError: When the arguments are not set correctly.
         """
         # TODO fix podman swagger cannot use this header!
         # headers = {"Content-type": "application/x-www-form-urlencoded"}
 
-        response = self.client.post(
-            "/images/load", data=data, headers={"Content-type": "application/x-tar"}
-        )
-        response.raise_for_status()
+        # Check that exactly one of the data or file_path is provided
+        if not data and not file_path:
+            raise PodmanError("The 'data' or 'file_path' parameter should be set.")
 
-        body = response.json()
-        for item in body["Names"]:
-            yield self.get(item)
+        if data and file_path:
+            raise PodmanError(
+                "Only one parameter should be set from 'data' and 'file_path' parameters."
+            )
+
+        post_data = data
+        if file_path:
+            # Convert to Path if file_path is a string
+            file_path_object = Path(file_path)
+            post_data = file_path_object.read_bytes()  # Read the tarball file as bytes
+
+        # Make the client request before entering the generator
+        response = self.client.post(
+            "/images/load", data=post_data, headers={"Content-type": "application/x-tar"}
+        )
+        response.raise_for_status()  # Catch any errors before proceeding
+
+        def _generator(body: dict) -> Generator[Image, None, None]:
+            # Iterate and yield images from response body
+            for item in body["Names"]:
+                yield self.get(item)
+
+        # Pass the response body to the generator
+        return _generator(response.json())
 
     def prune(
-        self, filters: Optional[Mapping[str, Any]] = None
-    ) -> Dict[Literal["ImagesDeleted", "SpaceReclaimed"], Any]:
+        self,
+        all: Optional[bool] = False,  # pylint: disable=redefined-builtin
+        external: Optional[bool] = False,
+        filters: Optional[Mapping[str, Any]] = None,
+    ) -> dict[Literal["ImagesDeleted", "SpaceReclaimed"], Any]:
         """Delete unused images.
 
         The Untagged keys will always be "".
 
         Args:
+            all: Remove all images not in use by containers, not just dangling ones.
+            external: Remove images even when they are used by external containers
+            (e.g, by build containers).
             filters: Qualify Images to prune. Available filters:
 
                 - dangling (bool): when true, only delete unused and untagged images.
+                - label: (dict): filter by label.
+                         Examples:
+                         filters={"label": {"key": "value"}}
+                         filters={"label!": {"key": "value"}}
                 - until (str): Delete images older than this timestamp.
 
         Raises:
             APIError: when service returns an error
         """
-        response = self.client.post(
-            "/images/prune", params={"filters": api.prepare_filters(filters)}
-        )
+
+        params = {
+            "all": all,
+            "external": external,
+            "filters": api.prepare_filters(filters),
+        }
+
+        response = self.client.post("/images/prune", params=params)
         response.raise_for_status()
 
-        deleted: List[Dict[str, str]] = []
-        error: List[str] = []
+        deleted: builtins.list[dict[str, str]] = []
+        error: builtins.list[str] = []
         reclaimed: int = 0
         # If the prune doesn't remove images, the API returns "null"
         # and it's interpreted as None (NoneType)
@@ -182,7 +229,7 @@ class ImagesManager(BuildMixin, Manager):
             "SpaceReclaimed": reclaimed,
         }
 
-    def prune_builds(self) -> Dict[Literal["CachesDeleted", "SpaceReclaimed"], Any]:
+    def prune_builds(self) -> dict[Literal["CachesDeleted", "SpaceReclaimed"], Any]:
         """Delete builder cache.
 
         Method included to complete API, the operation always returns empty
@@ -192,7 +239,7 @@ class ImagesManager(BuildMixin, Manager):
 
     def push(
         self, repository: str, tag: Optional[str] = None, **kwargs
-    ) -> Union[str, Iterator[Union[str, Dict[str, Any]]]]:
+    ) -> Union[str, Iterator[Union[str, dict[str, Any]]]]:
         """Push Image or repository to the registry.
 
         Args:
@@ -202,29 +249,37 @@ class ImagesManager(BuildMixin, Manager):
         Keyword Args:
             auth_config (Mapping[str, str]: Override configured credentials. Must include
                 username and password keys.
-            decode (bool): return data from server as Dict[str, Any]. Ignored unless stream=True.
+            decode (bool): return data from server as dict[str, Any]. Ignored unless stream=True.
             destination (str): alternate destination for image. (Podman only)
             stream (bool): return output as blocking generator. Default: False.
             tlsVerify (bool): Require TLS verification.
+            format (str): Manifest type (oci, v2s1, or v2s2) to use when pushing an image.
+                Default is manifest type of source, with fallbacks.
 
         Raises:
             APIError: when service returns an error
         """
-        auth_config: Optional[Dict[str, str]] = kwargs.get("auth_config")
+        auth_config: Optional[dict[str, str]] = kwargs.get("auth_config")
 
         headers = {
             # A base64url-encoded auth configuration
-            "X-Registry-Auth": encode_auth_header(auth_config) if auth_config else ""
+            "X-Registry-Auth": api.encode_auth_header(auth_config) if auth_config else ""
         }
 
         params = {
             "destination": kwargs.get("destination"),
             "tlsVerify": kwargs.get("tlsVerify"),
+            "format": kwargs.get("format"),
         }
 
+        stream = kwargs.get("stream", False)
+        decode = kwargs.get("decode", False)
+
         name = f'{repository}:{tag}' if tag else repository
         name = urllib.parse.quote_plus(name)
-        response = self.client.post(f"/images/{name}/push", params=params, headers=headers)
+        response = self.client.post(
+            f"/images/{name}/push", params=params, stream=stream, headers=headers
+        )
         response.raise_for_status(not_found=ImageNotFound)
 
         tag_count = 0 if tag is None else 1
@@ -239,8 +294,6 @@ class ImagesManager(BuildMixin, Manager):
             },
         ]
 
-        stream = kwargs.get("stream", False)
-        decode = kwargs.get("decode", False)
         if stream:
             return self._push_helper(decode, body)
 
@@ -251,8 +304,8 @@ class ImagesManager(BuildMixin, Manager):
 
     @staticmethod
     def _push_helper(
-        decode: bool, body: List[Dict[str, Any]]
-    ) -> Iterator[Union[str, Dict[str, Any]]]:
+        decode: bool, body: builtins.list[dict[str, Any]]
+    ) -> Iterator[Union[str, dict[str, Any]]]:
         """Helper needed to allow push() to return either a generator or a str."""
         for entry in body:
             if decode:
@@ -262,8 +315,12 @@ class ImagesManager(BuildMixin, Manager):
 
     # pylint: disable=too-many-locals,too-many-branches
     def pull(
-        self, repository: str, tag: Optional[str] = None, all_tags: bool = False, **kwargs
-    ) -> Union[Image, List[Image], Iterator[str]]:
+        self,
+        repository: str,
+        tag: Optional[str] = None,
+        all_tags: bool = False,
+        **kwargs,
+    ) -> Union[Image, builtins.list[Image], Iterator[str]]:
         """Request Podman service to pull image(s) from repository.
 
         Args:
@@ -275,7 +332,12 @@ class ImagesManager(BuildMixin, Manager):
             auth_config (Mapping[str, str]) – Override the credentials that are found in the
                 config for this request. auth_config should contain the username and password
                 keys to be valid.
+            compatMode (bool) – Return the same JSON payload as the Docker-compat endpoint.
+                Default: True.
+            decode (bool) – Decode the JSON data from the server into dicts.
+                Only applies with ``stream=True``
             platform (str) – Platform in the format os[/arch[/variant]]
+            policy (str) - Pull policy. "always" (default), "missing", "never", "newer"
             progress_bar (bool) - Display a progress bar with the image pull progress (uses
                 the compat endpoint). Default: False
             tls_verify (bool) - Require TLS verification. Default: True.
@@ -290,23 +352,24 @@ class ImagesManager(BuildMixin, Manager):
             APIError: when service returns an error
         """
         if tag is None or len(tag) == 0:
-            tokens = repository.split(":")
-            if len(tokens) == 2:
-                repository = tokens[0]
-                tag = tokens[1]
+            repository, parsed_tag = parse_repository(repository)
+            if parsed_tag is not None:
+                tag = parsed_tag
             else:
                 tag = "latest"
 
-        auth_config: Optional[Dict[str, str]] = kwargs.get("auth_config")
+        auth_config: Optional[dict[str, str]] = kwargs.get("auth_config")
 
         headers = {
             # A base64url-encoded auth configuration
-            "X-Registry-Auth": encode_auth_header(auth_config) if auth_config else ""
+            "X-Registry-Auth": api.encode_auth_header(auth_config) if auth_config else ""
         }
 
         params = {
+            "policy": kwargs.get("policy", "always"),
             "reference": repository,
-            "tlsVerify": kwargs.get("tls_verify"),
+            "tlsVerify": kwargs.get("tls_verify", True),
+            "compatMode": kwargs.get("compatMode", True),
         }
 
         if all_tags:
@@ -356,12 +419,12 @@ class ImagesManager(BuildMixin, Manager):
             return None
 
         if stream:
-            return response.iter_lines()
+            return self._stream_helper(response, decode=kwargs.get("decode"))
 
-        for item in response.iter_lines():
+        for item in reversed(list(response.iter_lines())):
             obj = json.loads(item)
             if all_tags and "images" in obj:
-                images: List[Image] = []
+                images: builtins.list[Image] = []
                 for name in obj["images"]:
                     images.append(self.get(name))
                 return images
@@ -406,7 +469,7 @@ class ImagesManager(BuildMixin, Manager):
         image: Union[Image, str],
         force: Optional[bool] = None,
         noprune: bool = False,  # pylint: disable=unused-argument
-    ) -> List[Dict[Literal["Deleted", "Untagged", "Errors", "ExitCode"], Union[str, int]]]:
+    ) -> builtins.list[dict[Literal["Deleted", "Untagged", "Errors", "ExitCode"], Union[str, int]]]:
         """Delete image from Podman service.
 
         Args:
@@ -425,7 +488,7 @@ class ImagesManager(BuildMixin, Manager):
         response.raise_for_status(not_found=ImageNotFound)
 
         body = response.json()
-        results: List[Dict[str, Union[int, str]]] = []
+        results: builtins.list[dict[str, Union[int, str]]] = []
         for key in ("Deleted", "Untagged", "Errors"):
             if key in body:
                 for element in body[key]:
@@ -433,14 +496,14 @@ class ImagesManager(BuildMixin, Manager):
         results.append({"ExitCode": body["ExitCode"]})
         return results
 
-    def search(self, term: str, **kwargs) -> List[Dict[str, Any]]:
+    def search(self, term: str, **kwargs) -> builtins.list[dict[str, Any]]:
         """Search Images on registries.
 
         Args:
             term: Used to target Image results.
 
         Keyword Args:
-            filters (Mapping[str, List[str]): Refine results of search. Available filters:
+            filters (Mapping[str, list[str]): Refine results of search. Available filters:
 
                 - is-automated (bool): Image build is automated.
                 - is-official (bool): Image build is owned by product provider.
@@ -493,3 +556,24 @@ class ImagesManager(BuildMixin, Manager):
         response = self.client.post(f"/images/scp/{source}", params=params)
         response.raise_for_status()
         return response.json()
+
+    def _stream_helper(self, response, decode=False):
+        """Generator for data coming from a chunked-encoded HTTP response."""
+
+        if response.raw._fp.chunked:
+            if decode:
+                yield from json_stream(self._stream_helper(response, False))
+            else:
+                reader = response.raw
+                while not reader.closed:
+                    # this read call will block until we get a chunk
+                    data = reader.read(1)
+                    if not data:
+                        break
+                    if reader._fp.chunk_left:
+                        data += reader.read(reader._fp.chunk_left)
+                    yield data
+        else:
+            # Response isn't chunked, meaning we probably
+            # encountered an error immediately
+            yield self._result(response, json=decode)
